/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: psect.c,v 1.2 2017-08-12 00:30:26-04 andrew Exp andrew $
 * 
 * Copyright (c) 2017 Andrew Turnquist
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * 
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "eval.h"
#include "opcodes.h"
#include "psect.h"
#include "util.h"

/*
Notes:

psect classes:
	CODE		code segment (program memory)
	EEDATA		EEPROM data
	CONFIG		Configuration memory (User ID space)
	COMMON		16 byte common memory
	BANK[0..n]	Banked memory pages
	LINEAR		Linear memory map
		N.B.: Must consider assignment overlap between
			BANKx and LINEAR

	psect   text0,local,class=CODE,merge=1,delta=2
	psect   maintext,global,class=CODE,merge=1,split=1,delta=2

Each section beginning with a PSECT directive will be processed separately
in pass 1 to generate a list of sections (either absolute or relative)
with size requirements to then be assigned memory space before beginning
pass 2. Absolute PSECTS (with an explicit ORG directive) are assigned
memory first. Relative PSECTS assigned memory where available.  Possible
allocation algorithms: First defined gets first available; largests
first; smallest first.

Defined symbols should have their PSECT associated with them.


Processor configuration data:
(PIC16F1825 example)
	codesize: 8192		; size of CODE
	eesize: 256		; size of EEDATA
	banks: 13		; Pages of banked memory, BANK[0..12]
	bankstart: 0x0020	; Start address of bank
	banksize: 80		; Size of BANKx
	banksize12: 48		; Size of BANK12
	ramsize: 1024		; Total RAM size (LINEAR size = ramsize-16)
	linearstart: 0x2000	; Start of linear addressing
	codestart: 0x8000	; Start of CODE in linear memory
	configstart: 0x8000	; Start of config memory in hexfile

BANKx from LINEAR: Bank = (Laddr - Lbase) / banksize
                   Addr = (Laddr - Lbase) % banksize + bankstart
LINEAR from BANKx: Laddr = (banknum * banksize) + (Addr - bankstart) + Lbase

LINEAR from CODE: Laddr = Lcodebase + Caddr
CODE from LINEAR: Caddr = Laddr - Lcodebase

Only CODE and EEDATA produce output for hexfile. Others are for
defining space used and assigning labels.

May define constant data in CODE, which is addressed via INDFx regs.
Should we have a STRCODE psect to store these separate from real code?



*/


#define PSTART(PSECT)	(((PSECT)->start == -1) ? 0 : (PSECT)->start)


/* These will become part of the processor config file. */
int codesize = 8192;
int eesize = 256;
int cmsize = 16;	// TODO: Check actual config mem size
int banks = 13;
int bankstart = 0x0020;
int banksize = 80;
int ramsize = 1024;
int linearstart = 0x2000;
int codestart   = 0x8000;

PsectList psects = NULL;
int lastpsect = 0;
struct psect * curpsect = NULL;


int psStrToClass(const char * str)
{
  char lstr[64];

  strncpy(lstr, str, 63);
  lowercase(lstr);
  if (!strcmp(lstr, "code"))   return PS_CODE;
  if (!strcmp(lstr, "eedata")) return PS_EEDATA;
  if (!strcmp(lstr, "config")) return PS_CONFIG;
  if (!strcmp(lstr, "common")) return PS_COMMON;
  if (!strcmp(lstr, "linear")) return PS_LINEAR;
  if (!strncmp(lstr, "bank", 4)) {
    if (!isdigit(lstr[4])) {
      message(MSG_ERROR, "psect: BANK number not specified");
      exit(1);
    }
    return PS_BANK0 + atoi(&(lstr[4]));
  }
  message(MSG_ERROR, "psect: class '%s' unknown", str);
  return 0;
}


/*
 * Address conversion utility functions.
 */

int psLinear2Bank(const int laddr)
{
  if (laddr < linearstart) return EVAL_FAIL;
  if (laddr >= codestart) return EVAL_FAIL;
  return (laddr - linearstart) / banksize;
}


int psLinear2BankAddr(const int laddr)
{
  if (laddr < linearstart) return EVAL_FAIL;
  if (laddr >= codestart) return EVAL_FAIL;
  return (laddr - linearstart) % banksize + bankstart;
}


int psBank2Linear(const int bank, const int baddr)
{
  return (bank * banksize) + (baddr - bankstart) + linearstart;
}


int psCode2Linear(const int caddr)
{
  return codestart + caddr;
}


int psLinear2Code(const int laddr)
{
  if (laddr < codestart) return EVAL_FAIL;
  return laddr - codestart;
}


/*
 * Find a psect by name.  Returns psect structure or NULL if not found.
 */

struct psect * psFind(const char *name)
{
  PsectList psl = psects;

  while (psl != NULL) {
    if (!strcmp((psl->p)->pname, name)) return psl->p;
    psl = psl->next;
  }
  MESSAGE(MSG_INFO, "psect %s not found", name);
  return NULL;
}


/*
 * Find a psect by id number.  Returns psect struct or NULL if not found.
 */

struct psect * psFindNum(const int pnum)
{
  PsectList psl = psects;

  while (psl != NULL) {
    if ((psl->p)->pnum == pnum) return psl->p;
    psl = psl->next;
  }
  MESSAGE(MSG_INFO, "psect %s not found", pnum);
  return NULL;
}


/*
 * Find next absolute or relative psect in list.
 */

PsectList psNext(PsectList psp, const int absolute)
{
  PsectList psl;

  psl = psp ? psp->next : psects;

  if (psl == NULL) return NULL;

  while (psl) {
    if (absolute && ((psl->p)->start >= 0)) return psl;
    if (!absolute && ((psl->p)->start < 0)) return psl;
    psl = psl->next;
  }
  return NULL;
}


/*
 * Find next psect matching 'class'.
 */

PsectList psNextInClass(PsectList psp, const int class)
{
  PsectList psl = psp ? psp->next : psects;

  if (psl == NULL) return NULL;

  while (psl) {
    if ((psl->p)->class == class) return psl;
    psl = psl->next;
  }
  return NULL;
}


int psFindFree(int *map, int mapsize, int wanted)
{
  int i,j;

  i=0;
  while (i<mapsize) {
    if (map[i] == 0) {
      /* Candidate memory slot. Check for 'wanted' bytes/words */
      for (j=0; (j<wanted) && (map[i+j] == 0); j++);

      /* If we found all 'wanted' bytes/words, return index */
      if (j == wanted) return i;

      /* Otherwise, skip past where we stopped */
      i += j;
    } else i++;
  }  /* while */
  /* If we got here, we failed to find 'wanted' contiguous bytes/words */
  return -1;
}


int psFindFreeInBank(const int bank, int *map, int mapsize, int wanted)
{
  int i,j;
  int eob;

  i=psBank2Linear(bank, 0 + bankstart) - linearstart;
  eob = (i+banksize) < mapsize ? i+banksize : mapsize;
  while (i < eob) {
    if (map[i] == 0) {
      /* Candidate memory slot. Check for 'wanted' bytes/words */
      for (j=0; (j<wanted) && (map[i+j] == 0); j++);

      /* If we found all 'wanted' bytes/words, return index */
      if (j == wanted) return psLinear2BankAddr(linearstart + i) - bankstart;

      /* Otherwise, skip past where we stopped */
      i += j;
    } else i++;
  }  /* while */
  /* If we got here, we failed to find 'wanted' continuous bytes */
  return -1;
}

/*
 * Sort all psects and assign memory start locations.  To be
 * called after pass1(), so pass2() knows where to put code segments
 * (and EEPROM data).
 */

void psAssign(void)
{
  PsectList psl = NULL;
  int i;
  int s;		/* Absolute start index */
  int b;		/* Bank number */
  int a;		/* Address */
  struct psect *p;	/* The psect we're working with */
  int *codewords;
  int *eewords;
  int *cmwords;
  int *combytes;
  int *linbytes;	/* Linear memory map; all banked memory mapped here */

  codewords = (int *)malloc(codesize * sizeof(int));
  for (i=0; i<codesize; i++) codewords[i] = 0;
  eewords = (int *)malloc(eesize * sizeof(int));
  for (i=0; i<eesize; i++) eewords[i] = 0;
  cmwords = (int *)malloc(cmsize * sizeof(int));
  for (i=0; i<cmsize; i++) cmwords[i] = 0;
  combytes = (int *)malloc(16 * sizeof(int));
  for (i=0; i<16; i++) combytes[i] = 0;
  linbytes = (int *)malloc((ramsize - 16) * sizeof(int));
  for (i=0; i<(ramsize - 16); i++) linbytes[i] = 0;

  /* Assign absolutes first. */
  MESSAGE(MSG_INFO, "Assigning space for absolute psects");
  while ((psl = psNext(psl, 1))) {
    p = psl->p;

    if (p->class == PS_CODE) {
      for (i=(p->start); (i<(p->start + p->length)) && (i< codesize); i++) {
        if (codewords[i] > 0)
            message(MSG_WARNING, "Overlapping absolute CODE psects at 0x%04x", i);
        codewords[i] = p->pnum;
      }  /* for */
    }  /* class == CODE */

    if (p->class == PS_EEDATA) {
      for (i=(p->start);
           (i<(p->start + p->length)) && ((i - p->start) < eesize);
           i++) {
        if (eewords[i] > 0)
            message(MSG_WARNING, "Overlapping absolute EEDATA psects at 0x%04x", i);
        eewords[i] = p->pnum;
      }  /* for */
    }  /* class == EEDATA */

    if (p->class == PS_CONFIG) {
      for (i=(p->start);
           (i<(p->start + p->length)) && ((i - p->start) < cmsize);
           i++) {
        if (cmwords[i] > 0)
            message(MSG_WARNING, "Overlapping absolute CONFIG psects at 0x%04x", i);
        cmwords[i] = p->pnum;
      }  /* for */
    }  /* class == CONFIG */

    if (p->class == PS_COMMON) {
      /* Common memory is 16 bytes from 0x70 to 0x7f */
      for (i=(p->start);
           (i<(p->start + p->length)) && (i < 16);
           i++) {
        if (combytes[i] > 0)
            message(MSG_NOTICE, "Overlapping absolute COMMON psects at 0x%04x", i);
        combytes[i] = p->pnum;
      }  /* for */
    }  /* class == COMMON */

    if (p->class == PS_LINEAR) {
      /* Linear memory starts at 0x2000 */
      for (i=(p->start);
           (i<(p->start + p->length)) && (i < (ramsize - 16));
           i++) {
        if (linbytes[i] > 0)
            message(MSG_NOTICE, "Overlapping absolute LINEAR psects at 0x%04x", i);
        linbytes[i] = p->pnum;
      }  /* for */
    }  /* class == LINEAR */

    // PS_BANKx classes are mapped into LINEAR space for assignments
    if (p->class >= PS_BANK0) {
      b = (p->class - PS_BANK0);
      for (i=(p->start);
           (i < (p->start + p->length)) && (i < banksize);
           i++) {
        a = psBank2Linear(b, i + bankstart) - linearstart;
        if (linbytes[a] > 0)
            message(MSG_NOTICE,
                    "Overlapping absolute BANK/LINEAR psects at 0x%04x",
                    a);
        linbytes[a] = p->pnum;
      }  /* for */
    }  /* class == BANKx */

  }  /* while absolute psects */


  /*
   * Assign memory for relocatable psects
   */

  MESSAGE(MSG_INFO, "Assigning space for relocatable psects");
  psl = NULL;
  while ((psl = psNext(psl, 0))) {
    p = psl->p;
    MESSAGE(MSG_INFO, "Assigning space for psect '%s'", p->pname);

    if (p->class == PS_CODE) {
      if ((s = psFindFree(codewords, codesize, p->length)) == -1) {
        message(MSG_ERROR, "Failed to allocate %d bytes for psect %s",
                p->length, p->pname);
        exit(1);
      }
      MESSAGE(MSG_INFO, "Setting start address of '%s' to 0x%04x", p->pname, s);
      p->start = s;
      for (i=(p->start); (i<(p->start + p->length)) && (i< codesize); i++) {
        if (codewords[i] > 0)
            message(MSG_CRITICAL, "Overlapping relocated CODE psects at 0x%04x. Assembler error?", i);
        codewords[i] = p->pnum;
      }  /* for */
    }  /* class == CODE */

    if (p->class == PS_EEDATA) {
      if ((s = psFindFree(eewords, eesize, p->length)) == -1) {
        message(MSG_ERROR, "Failed to allocate %d bytes for psect %s",
                p->length, p->pname);
        exit(1);
      }
      MESSAGE(MSG_INFO, "Setting start address of '%s' to 0x%04x", p->pname, s);
      p->start = s;
      for (i=(p->start); (i<(p->start + p->length)) && (i< eesize); i++) {
        if (eewords[i] > 0)
            message(MSG_CRITICAL, "Overlapping relocated EEDATA psects at 0x%04x. Assembler error?", i);
        eewords[i] = p->pnum;
      }  /* for */
    }  /* class == EEDATA */

    if (p->class == PS_CONFIG) {
      if ((s = psFindFree(cmwords, cmsize, p->length)) == -1) {
        message(MSG_ERROR, "Failed to allocate %d bytes for psect %s",
                p->length, p->pname);
        exit(1);
      }
      MESSAGE(MSG_INFO, "Setting start address of '%s' to 0x%04x", p->pname, s);
      p->start = s;
      for (i=(p->start); (i<(p->start + p->length)) && (i< cmsize); i++) {
        if (cmwords[i] > 0)
            message(MSG_CRITICAL, "Overlapping relocated CONFIG psects at 0x%04x. Assembler error?", i);
        cmwords[i] = p->pnum;
      }  /* for */
    }  /* class == CONFIG */

    if (p->class == PS_COMMON) {
      if ((s = psFindFree(combytes, 16, p->length)) == -1) {
        message(MSG_ERROR, "Failed to allocate %d bytes for psect %s",
                p->length, p->pname);
        exit(1);
      }
      MESSAGE(MSG_INFO, "Setting start address of '%s' to 0x%04x", p->pname, s);
      p->start = s;
      for (i=(p->start); (i<(p->start + p->length)) && (i< 16); i++) {
        if (combytes[i] > 0)
            message(MSG_CRITICAL, "Overlapping relocated COMMON psects at 0x%04x. Assembler error?", i);
        combytes[i] = p->pnum;
      }  /* for */
    }  /* class == COMMON */

    if (p->class == PS_LINEAR) {
      if ((s = psFindFree(linbytes, (ramsize - 16), p->length)) == -1) {
        message(MSG_ERROR, "Failed to allocate %d bytes for psect %s",
                p->length, p->pname);
        exit(1);
      }
      MESSAGE(MSG_INFO, "Setting start address of '%s' to 0x%04x", p->pname, s);
      p->start = s;
      for (i=(p->start); (i<(p->start + p->length)) && (i< (ramsize - 16)); i++) {
        if (linbytes[i] > 0)
            message(MSG_CRITICAL, "Overlapping relocated LINEAR psects at 0x%04x. Assembler error?", i);
        linbytes[i] = p->pnum;
      }  /* for */
    }  /* class == LINEAR */

    if (p->class >= PS_BANK0) {
      b = p->class - PS_BANK0;
      if ((s = psFindFreeInBank(b, linbytes, (ramsize - 16), p->length)) == -1)
      {
        message(MSG_ERROR, "Failed to allocate %d bytes for psect %s",
                p->length, p->pname);
        exit(1);
      }
      MESSAGE(MSG_INFO, "Setting start address of '%s' to 0x%04x", p->pname, s);
      p->start = s;
      for (i=(p->start); (i<(p->start + p->length)) && (i< banksize); i++) {
        a = psBank2Linear(b, i + bankstart) - linearstart;
        if (linbytes[a] > 0)
            message(MSG_CRITICAL, "Overlapping relocated BANK/LINEAR psects at 0x%04x. Assembler error?", i);
        linbytes[a] = p->pnum;
      }  /* for */
    }  /* class == PS_BANKx */

  }  /* while (psNext) relocatable */


  free(linbytes);
  free(combytes);
  free(cmwords);
  free(eewords);
  free(codewords);
}


/*
 * Create a new psect and add to list.
 */

struct psect * psNew(const char *name, const int class, const int start)
{
  struct psect *ps;
  PsectList psl;

  MESSAGE(MSG_VERBOSE, "Creating new psect %s", name);

  psl = (PsectList)malloc(sizeof(struct psect_list));
  psl->next = psects;
  ps = psl->p = (struct psect *)malloc(sizeof(struct psect));
  ps->pnum = ++lastpsect;
  ps->pname = strdup(name);
  ps->class = class;
  ps->start = start;
  ps->length = 0;
  ps->lc = 0;
  psects = psl;
  return ps;
}


void psDefine(const char *name, const char *class, const char *start)
{
  struct psect *ps;
  int s;

  if ((ps = psFind(name)) != NULL) {
    if (!pass) message(MSG_WARNING, "Attempt to redefine psect %s, ignoring new definition", name);
    curpsect = ps;
    return;
  }

  // sscanf(start, "%i", &s);
  s = evalString(start);
  curpsect = psNew(name, psStrToClass(class), s);
  MESSAGE(MSG_INFO, "Defined new psect %s", name);
}


void psDoDirective(char * params)
{
  int numtoks = 0;
  char * tok;
  char * tokptr;
  char * subptr;
  char * param;
  char * value;
  char *name = NULL;
  char *class = NULL;
  char *start = NULL;

  while ((tok = strtok_r(numtoks ? NULL : params, ",", &tokptr))) {
    MESSAGE(MSG_DEBUG, "Got token '%s'", tok);
    if (!numtoks) name = tok; else {
      param=strtok_r(tok,  "=", &subptr);
      value=strtok_r(NULL, "=", &subptr);
      if (!strcmp(param, "class")) class = value;
      if (!strcmp(param, "abs"))   start = value;
    }
    numtoks++;
  }
  if (class) {
    psDefine(name, class, start ? start : "-1");
  } else {
    MESSAGE(MSG_VERBOSE, "Using existing psect '%s'", name);
    curpsect = psFind(name);
    if (curpsect == NULL) {
      message(MSG_NOTICE, "psect '%s' not found, reverting to '%s'",
              name, DEFAULT_PSECT);
      curpsect = psFind(DEFAULT_PSECT);
    }  /* default psect */
  }  /* else class */

  if (lstp) {
    fprintf(lstp, "%04x", curpsect->start);
    fprintf(lstp, "\t\tpsect\t%s\n", params);
  }
}


void psInit(const int nbanks)
{
  banks = nbanks;
  psects = NULL;
  curpsect = psNew(DEFAULT_PSECT, PS_CODE, 0);

}


/* Destroy the entire psect list */
void psDestroy(void)
{
  PsectList psl;

  while (psects) {
    if (psects->p) free(psects->p);
    psl = psects;
    psects = psects->next;
    free(psl);
  }
}


/* Reset all psect location counters to zero */
void psResetAll(void)
{
  PsectList psl = psects;

  while (psl) {
    (psl->p)->lc = 0;
    psl = psl->next;
  }  /* while */
}


/* Returns true (1) if memory class gets written to hexfile */
int psIsHexMem(int class)
{
  if (class == PS_CODE) return 1;
  if (class == PS_EEDATA) return 1;
  if (class == PS_CONFIG) return 1;
  return 0;
}


/*
 * Get the base address for the given psect
 */

int psAddrBase(const int pnum)
{
  struct psect * p = psFindNum(pnum);
  int b;

  switch (p->class) {
    case PS_CODE: return 0x0000 + PSTART(p); break;   /* For code labels; 0x8000 if */
					  /* read via FSRx              */
    case PS_EEDATA: return 0x0000 + PSTART(p); break; /* Only accessible by EE* regs */
    case PS_CONFIG: return 0x0000 + PSTART(p); break; /* Only accessible by EE* regs */
    case PS_COMMON: return 0x0070 + PSTART(p);   break; /* Any bank */
    case PS_LINEAR: return 0x2000 + PSTART(p); break; /* Via FSRx */
    default:
      if (p->class >= PS_BANK0) {
        b = (p->class - PS_BANK0);
        return (b * 0x0080) + bankstart + PSTART(p);
      } else {
        return 0x0000 + PSTART(p);  /* Unknown class */
      }
      break;
  }
}


void psUsageSummary(FILE *pfp)
{
  int used;
  int cused;
  int lused;
  int bank;
  PsectList psl;

  fprintf(pfp, "; Memory usage by class:\n");
  
  used = 0;
  psl = NULL;
  while ((psl = psNextInClass(psl, PS_CODE)) != NULL) {
    used += (psl->p)->length;
  }
  fprintf(pfp, "; CODE    % 5d/% 5d = % 3d%%\n", used, codesize, (used * 100) / codesize);

  used = 0;
  psl = NULL;
  while ((psl = psNextInClass(psl, PS_EEDATA)) != NULL) {
    used += (psl->p)->length;
  }
  fprintf(pfp, "; EEDATA  % 5d/% 5d = % 3d%%\n", used, eesize, (used * 100) / eesize);

  cused = 0;
  psl = NULL;
  while ((psl = psNextInClass(psl, PS_COMMON)) != NULL) {
    cused += (psl->p)->length;
  }
  fprintf(pfp, "; COMMON  % 5d/% 5d = % 3d%%\n", cused, 16, (cused * 100) / 16);

  lused = 0;
  for (bank=0; bank<banks; bank++) {
    used = 0;
    psl = NULL;
    while ((psl = psNextInClass(psl, PS_BANK0 + bank)) != NULL) {
      used += (psl->p)->length;
    }
    fprintf(pfp, "; BANK%d%s  % 5d/% 5d = % 3d%%\n", bank, (bank < 10) ? " " : "", used, banksize, (used * 100) / banksize);
    lused += used;
  }

  used = 0;
  psl = NULL;
  while ((psl = psNextInClass(psl, PS_LINEAR)) != NULL) {
    used += (psl->p)->length;
  }
  fprintf(pfp, "; LINEAR  % 5d/% 5d = % 3d%%\n", used, ramsize - 16, (used * 100) / (ramsize - 16));
  lused += used;
  fprintf(pfp, "; TLINEAR % 5d/% 5d = % 3d%%\n", lused, ramsize - 16,
          (lused * 100) / (ramsize - 16));
  fprintf(pfp, "; TSRAM   % 5d/% 5d = % 3d%%\n", lused+cused, ramsize,
          ((lused+cused) * 100) / ramsize);
}


void psDump(FILE *pfp)
{
  PsectList psl = psects;
  struct psect * p;
  char cstr[16];

  while (psl) {
    p = psl->p;
    switch (p->class) {
      case PS_CODE:   sprintf(cstr, "CODE"); break;
      case PS_EEDATA: sprintf(cstr, "EEDATA"); break;
      case PS_CONFIG: sprintf(cstr, "CONFIG"); break;
      case PS_COMMON: sprintf(cstr, "COMMON"); break;
      case PS_LINEAR: sprintf(cstr, "LINEAR"); break;
      default:
        if (p->class >= PS_BANK0) {
          sprintf(cstr, "BANK%d", (p->class - PS_BANK0)); break;
        } else {
          sprintf(cstr, "%d", p->class); break;
        }
    }
    fprintf(pfp, "%s %d %s 0x%04x 0x%04x(%d)\n", p->pname, p->pnum, cstr,
        p->start, p->length, p->length);

    psl = psl->next;
  }
  fprintf(pfp, "\n");
  psUsageSummary(pfp);
}

