/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: symtab.c,v 1.2 2017-08-12 00:30:26-04 andrew Exp andrew $
 * 
 * Copyright (c) 2017 Andrew Turnquist
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * 
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "eval.h"
#include "lc.h"
#include "symtab.h"
#include "opcodes.h"
#include "psect.h"
#include "util.h"



/* Hash of SYMTAB_SIZE symlists */
struct symlist *symtabs[SYMTAB_SIZE];


void symInit(void)
{
  int i;

  for (i=0; i<SYMTAB_SIZE; i++) symtabs[i] = NULL;
}


void symDestroy(void)
{
  int i;
  struct symlist *sp;

  for (i=0; i<SYMTAB_SIZE; i++) {
    while (symtabs[i] != NULL) {
      sp = (symtabs[i])->next;
      free((symtabs[i])->label);
      free(symtabs[i]);
      symtabs[i] = sp;
    }
  }
}


void symAdd(const char *buf, const int loc, unsigned int f)
{
  int i;
  struct symlist *sent;

  if (!buf[0]) {
    fprintf(stderr, "WARNING: Attempt to add null symbol. Ignoring...\n");
    return;
  }

  // fprintf(stderr, "%04x symAdd: Adding symbol '%s' = %04x, flags = %04x\n",
  //     lc, buf, loc, f);
  // fprintf(stderr, "     Pearson hash = %02x\n", Pearson_hash((void *)buf, strlen(buf)));
  i = Pearson_hash((void *)buf, strlen(buf)) % SYMTAB_SIZE;
  // i = (buf[0] % SYMTAB_SIZE);
  sent = (struct symlist *)malloc(sizeof(struct symlist));
  sent->label = strdup(buf);
  sent->val = loc;
  sent->psect = curpsect->pnum;
  sent->flags = f;
  sent->next = symtabs[i];
  symtabs[i] = sent;
}


int symLookup(const char *lbl)
{
  int i;
  struct symlist *sp;

  i = Pearson_hash((void *)lbl, strlen(lbl)) % SYMTAB_SIZE;
  // i = (lbl[0] % SYMTAB_SIZE);
  sp = symtabs[i];
  while (sp) {
    if (!strcmp(sp->label, lbl)) {
      return ((sp->flags & SF_ABS) ? 0 : psAddrBase(sp->psect)) + sp->val;
    }
    sp = sp->next;
  }
  return EVAL_FAIL;
}


void symDump(FILE *lfp)
{
  int i;
  int n;
  int nmax = 0;
  struct symlist *sp;

  for (i=0; i<SYMTAB_SIZE; i++) {
    sp = symtabs[i];
    n = 0;
    while (sp != NULL) {
      fprintf(lfp, "%s@%d %04x %04x\n", sp->label, sp->psect, sp->val, sp->flags);
      sp = sp->next;
      n++;
    }
    fprintf(lfp, "; Bucket %d list depth: %d\n", i, n);
    if (n > nmax) nmax = n;
  }
  fprintf(lfp, "; Max list depth: %d\n", nmax);
}


int symResolveAddr(const char *op)
{
  int opa;
  char alabel[64];

  // Evaluate the string. If not error condition, return value.
  opa = evalString(op);
  if (opa != EVAL_FAIL) return opa;

  // Really, the rest of this should now be redundant.
  MESSAGE(MSG_VERBOSE, "symResolveAddr: evalString(\"%s\") failed. Why?", op);

  // Numeric operand
  //
  if ( isdigit(op[0]) || ( (tolower(op[0]) == 'b') && (op[1] == '\''))) {
    // sscanf(op, "%i", &opa);
    opa = parseNumber(op);
    if (opa != EVAL_FAIL) return opa;
  }

  // Label operand
  //
  int i=0;
  while (op[i] && (op[i]!=' ') && (op[i]!='\t')) {
    alabel[i] = op[i];
    i++;
  }
  alabel[i] = '\0';

  // Special "current location counter" symbol '$'
  //
  if (!strcmp(alabel, "$")) return lc;

  opa = symLookup(alabel);
  if (opa == EVAL_FAIL) {
    return EVAL_FAIL;
  }
  return opa;
}


int symResolveLiteral(const char *op)
{
  int opa;
  char alabel[64];

  MESSAGE(MSG_DEBUG, "Entering symResolveLiteral; operand = '%s'", op);

  // Evaluate the string. If not error condition, return value.
  opa = evalString(op);
  if (opa != EVAL_FAIL) return opa;

  message(MSG_NOTICE, "evalString() failed resolving");

  // Numeric operand
  if ( isdigit(op[0]) || ( (tolower(op[0]) == 'b') && (op[1] == '\''))) {
    // sscanf(op, "%i", &opa);
    opa = parseNumber(op);
    if (opa != EVAL_FAIL) return opa;
  }

  // Constant operand
  int i=0;
  while (op[i] && (op[i]!=' ') && (op[i]!='\t')) {
    alabel[i] = op[i];
    i++;
  }
  alabel[i] = '\0';

  if (!strcmp(alabel, "$")) return lc;
  MESSAGE(MSG_DEBUG, "symResolveLiteral: Looking up symbol %s\n", alabel);
  opa = symLookup(alabel);
  // opa = 0;
  if (opa == EVAL_FAIL) {
    // TODO: Should this just return EVAL_FAIL instead, for fwd refs?
    message(pass ? MSG_ERROR : MSG_NOTICE, "symResolveLiteral: symbol '%s' not found",
        alabel);
    if (pass) exit(1); else return EVAL_FAIL;
  } else {
    MESSAGE(MSG_INFO, "symResolveLiteral: Found symbol '%s' = %04x\n", alabel, opa);
  }
  return opa;
}

