; Program: Uart2Spi
;  Author: Andrew Turnquist <andrew@turnquist.name>
; Version: 1.0
; Revision: $Id:$
;
; Description: Translates UART to/from SPI host.



	include pic16f1825.inc

	__config1	CONFIG_FCMEN_OFF & CONFIG_IESO_OFF & CONFIG_FOSC_INTOSC & 0xf9ef
	__config2	0x1eff


	macro	clear	; Clear (zero) register
	banksel	$1
	clrf	$1
	endm

	macro	movlf
	movlw	$1	; Load literal into W
	banksel	$2
	movwf	$2	; Save W into register
	endm

	macro	brz	; Branch if not zero
	btfsc	STATUS,Z
	bra	$1
	endm

	macro	brnz	; Branch if not zero
	btfss	STATUS,Z
	bra	$1
	endm

	macro	shiftl	; shiftl <register>,<bits>
	banksel	$1		; Make sure we've selected right bank
	movlw	$2		; Load bits into W
@slloop:
	lslf	$1,f		; Shift
	decfsz	WREG,f		; and decrement bits
	bra	@slloop
	endm

	macro	shiftr	; shiftr <register>,<bits>
	banksel	$1		; Make sure we've selected right bank
	movlw	$2		; Load bits into W
@slloop:
	lsrf	$1,f		; Shift
	decfsz	WREG,f		; and decrement bits
	bra	@slloop		; loop if bits > 0
	endm

	macro	di
	bcf	INTCON,GIE
	endm

	macro	ei
	bsf	INTCON,GIE
	endm

	macro fgetc
	di
	call fgetc_i
	ei
	endm

	; Sets Z flag if RB_QSIZE is empty, clear otherwise
	macro	fready
	banksel	$1
	movf	RB_QSIZE,f	; Test if queue is empty
	endm
	
	macro kbhit
	banksel RB_CMD
	movf	RB_QSIZE,f	; Test if queue is empty
	endm

	macro getch
	movlw	RB_CMD
	fgetc
	endm


DRDY		EQU	LATA1
SENDCMD		EQU	RA2

RB_RX		EQU	1
RB_TX		EQU	2
RB_CMD		EQU	3

RB_SIZE		EQU	64
RB_CTS_THRESH	EQU	48

; Common memory area (16 bytes)
;TEMP0		EQU	0x070	; Temporary holding register
;BAUD		EQU	0x071	; Baud rate index

	psect common,class=COMMON
temp0		db	0
baud		db	0

; These exist in banks 1, 2, and 3
RB		EQU	0x020	; 0x0a0, 0x120, 0x1a0
RB_HEAD		EQU	0x060	; 0x0e0, 0x160, 0x1e0
RB_TAIL		EQU	0x061	; 0x0e1, 0x161, 0x1e1
RB_QSIZE	EQU	0x062	; 0x0e2, 0x162, 0x1e2
RB_OVERRUN	EQU	0x063	; 0x0e3, 0x163, 0x1e3


B300		EQU	0
B1200		EQU	1
B2400		EQU	2
B4800		EQU	3
B9600		EQU	4
B19200		EQU	5
B38400		EQU	6
B57600		EQU	7


	psect default
	ORG	0x0000
_start:
	goto_near start


	psect isr,class=CODE,abs=0x0004
	org	0x0000
;	ORG	0x0004
isr:
	clrwdt
	banksel	PIE1		; if SSP1IE
	btfss	PIE1,SSP1IE
	bra	isr_rx
	banksel	PIR1		; && SSP1IF
	btfss	PIR1,SSP1IF
	bra	isr_rx
				; Handle SPI In/Out
	movlf	(SSP1BUF & 0xff00) >> 8,FSR1H	; Ready SSP1BUF for fputc
	movlf	(SSP1BUF & 0x00ff),FSR1L
				; TODO: Check CMD pin and use RB_CMD if set
	movlw	RB_TX		; Set output queue to UART TX
	banksel	PORTA		; If SENDCMD asserted ...
	btfsc	PORTA,SENDCMD
	movlw	RB_CMD		; ... change queue to command queue.
	call_near fputc_i	; Put the character
isr_s1:				; Force banksel code generation
	banksel	PIR1		; Clear the interrupt
	bcf	PIR1,SSP1IF
	
	movlb	RB_RX		; Is the queue empty?
	movf	RB_QSIZE,f
	btfss	STATUS,Z
	bra	isr_sput	; No. Skip over DRDY change
	banksel	LATA		; Deassert DRDY
	bcf	LATA,DRDY
	bra	isr_rx		; And we're done with SPI
isr_sput:
	movlw	RB_RX		; Get next byte from RX queue
	call_near fgetc_i
isr_s2:				; Force banksel code generation
	banksel	SSP1BUF		; Write the byte to SSP1BUF
	movwf	SSP1BUF		;   ... DRDY should already be high
isr_rx:
	banksel	PIE1		; if RCIE
	btfss	PIE1,RCIE
	bra	isr_tx
	banksel	PIR1		; && RCIF
	btfss	PIR1,RCIF
	bra	isr_tx
	movlw	0x01		; Point 'c' (INDF1) to RCREG
	movwf	FSR1H
	movlw	0x99
	movwf	FSR1L
	movlw	RB_RX		; 'stream' = RB_RX
	call_near fputc_i
isr_tx:
	banksel	PIE1		; if TXIE
	btfss	PIE1,TXIE
	retfie
	banksel	PIR1		; && RCIF
	btfss	PIR1,TXIF
	retfie
	fready	RB_TX		; Is queue empty?
	btfss	STATUS,Z
	bra	isr_tx_e	; Yes
	movlw	RB_TX		; No: write next char to TXREG
	call_near fgetc_i
	banksel	TXREG
	movwf	TXREG
	retfie
isr_tx_e:
	; RB_TX empty
	bcf	PIE1,TXIE	; Disable interrupt until something in queue
	retfie

	psect mainprog,class=CODE

; Init the ring buffers. No params, no return value.
initrbs:
	movlw	0x00
	movlb	4
irbs_loop:
	clrf	RB_HEAD
	clrf	RB_TAIL
	clrf	RB_QSIZE
	decfsz	BSR,f
	bra	irbs_loop
	return

; fgetc without interrupt twiddling. Called by fgetc or direct from ISR.
; stream in W, return value in W
;
fgetc_i:
	movwf	BSR		; Select buffer page
				; If we need 'stream' again, its it BSR

	; if (!rb_qsize) return 0;
	movf	RB_QSIZE,f	; Test RB_QSIZE
	btfss	STATUS,Z	; Skip if nonzero
	retlw	0		; If RB_QSIZE == 0, return 0

	; ch = rb[rb_tail]
	lsrf	BSR,W		; Get high bits of stream
	movwf	FSR0H		; Write high bits
	movf	RB_TAIL,W	; Get tail value
	addlw	RB		; Add base of ring buffer
	btfsc	BSR,0		; Set bit 7 to BSR:0
	bsf	WREG,7
	movwf	FSR0L		; Write low bits
				; character is in INDF0 now

	; rb_tail++;
	incf	RB_TAIL,f

	; if (rb_tail == RB_SIZE) rb_tail = 0
	movlw	RB_SIZE		; Get size
	subwf	RB_TAIL,W	; Subtract tail
	btfsc	STATUS,Z	; If result is zero reset TB_TAIL
	clrf	RB_TAIL

	; if (rb_qsize) rb_qsize--
	movf	RB_QSIZE,f	; Test RB_QSIZE
	btfss	STATUS,Z	; If not zero ...
	decf	RB_QSIZE,f	;   decrement it

	movf	INDF0,W		; Get char into W
	return

;fgetc:
;	di
;	call fgetc_i
;	ei
;	return

; 'stream' in W
starttx:
	subwf	RB_RX,W		; Subtract RB_RX
	btfss	STATUS,Z	; If result is zero, do RB_RX case
	bra	st_tx		; If W != RB_RX, try next option
	btfss	SSP1STAT,BF
	return
				; Should the interrupt be cleared?
	banksel	PIR1
	bcf	PIR1,SSP1IF	; Clear any SSP1 interrupt
	movlw	RB_RX		; Get character from RB_RX buffer
	call_near fgetc_i
	movwf	SSP1BUF		; Write it to SPI port
	banksel	LATA
	bsf	LATA,DRDY
	return
st_tx:
	addwf	RB_RX,W		; Restore W
	subwf	RB_TX,W		;  Try RB_TX
	btfss	STATUS,Z	; If zero, do RB_TX case
	bra	st_cmd		;   else try next option
	btfss	PIR1,TXIF	; If not TXIF
	return			;   return
	movlw	RB_TX		; Get character from RB_TX buffer
	call_near fgetc_i
	movwf	TXREG		; Write it to the UART
	banksel	PIE1		; Enable TXIE
	bsf	PIE1,TXIE
	return
st_cmd:
	addwf	RB_TX,W		; Restore W
	; Nothing to do for CMD queue or otherwise
	return


; 'c' in INDF1, 'stream' in W
; returns 'c' in W (Need I bother? It'll still be in INDF1, probably)
;
fputc_i:
	movwf	BSR		; Select page for stream
	movlw	(RB_SIZE - 2)	; rb_qsize - (RB_SIZE - 2)
	subwf	RB_QSIZE,W
; 61 >= 62 F --> 62 > 61 T C=0
; 62 >= 62 T --> 62 > 62 F C=1 Z=1
; 63 >= 62 T --> 62 > 63 F C=1 Z=0
	btfss	STATUS,C	; if (RB_SIZE - 2) > rb_qsize, skip
	bra	fp_else		;   else goto else part
	movlw	1		; rb_overrun = 1
	movwf	RB_OVERRUN
	retlw	255		; Return error
fp_else:
	; rb[rb_head] = c
	lsrf	BSR,W		; Get high bits of stream
	movwf	FSR0H		; Write high bits
	movf	RB_HEAD,W	; Get tail value
	addlw	RB		; Add base of ring buffer
	btfsc	BSR,0		; Set bit 7 to BSR:0
	bsf	WREG,7
	movwf	FSR0L		; Write low bits
	movf	INDF1,W		; Copy 'c' to INDF0
	movwf	INDF0

	; rb_head++;
	incf	RB_HEAD,f

	; if (rb_head == RB_SIZE) rb_head = 0
	movlw	RB_SIZE
	subwf	RB_HEAD,W
	btfsc	STATUS,Z
	clrf	RB_HEAD

	; rb_qsize++;
	incf	RB_QSIZE,f

	; if (rb_qsize == 1) starttx(stream);
	decf	RB_QSIZE,W
	btfss	STATUS,Z
	bra	fp_ret
	movf	BSR,W
	call	starttx
fp_ret:
	movf	INDF1,W
	return



start:
	clrwdt
	movlf	0x72,OSCCON	; 8MHz, no PLL, Int Osc
	clear	OPTION_REG
	clear	PORTA		; Init PORTA pins before activating
	clear	LATA
	movlf	0b00111101,TRISA	; PORTA inputs RA5,RA4,RA3,RA2,RA0
	movlf	0b00101011,TRISC	; PORTC inputs RC5,RC3,RC1,RC0
	clear	WPUA		; Zero WPUA
	clear	WPUC		; Zero WPUC
	banksel	ANSELC		; and ANSELC
	clrf	ANSELC		; and ANSELC
; Comparators default to off on 1825
;	clear	CM1CON0		; Turn off comparator (CMxCON0:7)
;	clear	CM2CON0

; Init UART
	movlf	0x26,TXSTA	; TX enabled, BRGH
	movlf	0x90,RCSTA	; UART enabled, RX enabled
	movlf	0x08,BAUDCON	; BRG16
	movlw	B9600
	call_near	setbaud
;	movlf	(207 & 0xff00) >> 8,SPBRGH	; BRG = 207 (9600bps)
;	movlf	(207 & 0x00ff),SPBRGL

; Init SPI
	movlf	0x24,SSP1CON1	; SPI Slave mode w/SS1, on
	movlf	0x10,SSP1CON3	; Ignore buffer overflows

; Init Interrupts
	movlf	0x40,INTCON	; PEIE enabled, GIE off for now
	movlf	0x28,PIE1	; Enable RCIE,SSP1IE; no TXIE until
				;   we have data to send
; Init WDT
	clrwdt
	movlf	0x15,WDTCON	; Watchdog enabled, 1s interval

	call initrbs
	ei
mainloop:
	clrwdt
	banksel	PORTC
	bsf	PORTC,RC0	; Set PORTC0 bit
	nop
	nop
	nop
	nop
	bcf	PORTC,RC0	; Clear PORTC0 bit
	kbhit			; If no bytes in queue
	brz	mainloop	; ... loop
				; Process next byte
	getch			; Get the byte into W
	movwf	temp0		; Stash the byte
	sublw	0x42		; Is it a 'B' (baud set)?
	brnz	main_end_switch	; ... no. Skip over B code

main_b_wait:
	kbhit			; Wait for next character
	brz main_b_wait		; We could clear WDT here, but
				;   the next character shouldn't
				;   take more than one second to arrive
	getch
	movwf	temp0
	sublw	0x30		; Transpose ASCII to binary
	call_near setbaud

main_end_switch:
	goto mainloop


	psect setbaud,class=CODE
	org	0x0000
;	org 	0x1f9
setbaud:
	banksel SPBRGL
	andlw	0x07
	movwf	baud
	lslf	WREG,W
	lslf	WREG,W
	lslf	WREG,W
	brw

SETBAUD_BASE	EQU 	$
	org	(SETBAUD_BASE + 8 * 0)		; W = 0, 300bps
	movlw	(6666 >> 8)
	movwf	SPBRGH
	movlw	(6666 & 0xff)
	movwf	SPBRGL
	return			; 5 instructions

	org	(SETBAUD_BASE + 8 * 1)		; W = 1, 1200bps
	movlw	(1666 >> 8)
	movwf	SPBRGH
	movlw	(1666 & 0xff)
	movwf	SPBRGL
	return

	org	(SETBAUD_BASE + 8 * 2)		; W = 2, 2400bps
	movlw	(832 >> 8)
	movwf	SPBRGH
	movlw	(832 & 0xff)
	movwf	SPBRGL
	return

	org	(SETBAUD_BASE + 8 * 3)		; W = 3, 4800bps
	movlw	(417 >> 8)
	movwf	SPBRGH
	movlw	(417 & 0xff)
	movwf	SPBRGL
	return

	org	(SETBAUD_BASE + 8 * 4)		; W = 4, 9600bps
	movlw	(207 >> 8)
	movwf	SPBRGH
	movlw	(207 & 0xff)
	movwf	SPBRGL
	return

	org	(SETBAUD_BASE + 8 * 5)		; W = 5, 19.2kbps
	movlw	(103 >> 8)
	movwf	SPBRGH
	movlw	(103 & 0xff)
	movwf	SPBRGL
	return

	org	(SETBAUD_BASE + 8 * 6)		; W = 6, 38.4kbps
	movlw	(52 >> 8)
	movwf	SPBRGH
	movlw	(52 & 0xff)
	movwf	SPBRGL
	return

	org	(SETBAUD_BASE + 8 * 7)		; W = 7, 57.6kbps
	movlw	(34 >> 8)
	movwf	SPBRGH
	movlw	(34 & 0xff)
	movwf	SPBRGL
	return

;	org	0x240
;
;	moviw	FSR0++
;	moviw	-2[FSR1]
;	movwi	--FSR0
;	movwi	2[FSR0]

;	org	0x400
end:
END_OF_CODE	EQU	$

;	END
